from octopus.core.function import Function
from octopus.core.basicblock import BasicBlock
from octopus.core.edge import (Edge,
                               EDGE_UNCONDITIONAL,
                               EDGE_CONDITIONAL_TRUE, EDGE_CONDITIONAL_FALSE,
                               EDGE_FALLTHROUGH, EDGE_CALL)
from octopus.analysis.cfg import CFG
from octopus.analysis.graph import CFGGraph
from octopus.arch.wasm.analyzer import WasmModuleAnalyzer
from octopus.arch.wasm.disassembler import WasmDisassembler
from octopus.arch.wasm.format import (format_func_name,
                                      format_bb_name)

from octopus.arch.wasm.wasm import _groups
from octopus.core.utils import bytecode_to_bytes
# for graph visualisation
from graphviz import Digraph
from logging import getLogger
logging = getLogger(__name__)


DESIGN_IMPORT = {'fillcolor': 'turquoise',
                 'shape': 'box',
                 'style': 'filled'}

DESIGN_EXPORT = {'fillcolor': 'grey',
                 'shape': 'box',
                 'style': 'filled'}


def enum_func(module_bytecode):
    ''' return a list of Function
        see:: octopus.core.function
    '''
    functions = list()
    analyzer = WasmModuleAnalyzer(module_bytecode)

    protos = analyzer.func_prototypes
    import_len = len(analyzer.imports_func)

    for idx, code in enumerate(analyzer.codes):
        # get corresponding function prototype
        name, param_str, return_str, _ = protos[import_len + idx]

        prefered_name = format_func_name(name, param_str, return_str)
        instructions = WasmDisassembler().disassemble(code)
        cur_function = Function(0, instructions[0], name=name, prefered_name=prefered_name)
        cur_function.instructions = instructions

        functions.append(cur_function)
    return functions


def enum_func_call_edges(functions, len_imports):
    ''' return a list of tuple with
        (index_func_node_from, index_func_node_to)
    '''
    call_edges = list()

    # iterate over functions
    for index, func in enumerate(functions):
        node_from = len_imports + index
        # iterates over instruction
        for inst in func.instructions:
            # detect if inst is a call instructions
            if inst.name == "call":  #is_call:
                logging.info('%s', inst.operand_interpretation)
                #if inst.name == "call":
                # only get the import_id
                node_to = int(inst.operand_interpretation.split(' ')[1])
                # The `call_indirect` operator takes a list of function arguments and as the last operand the index into the table.
                #elif inst.name == "call_indirect":
                # the last operand is the index on the table
                #print(inst.operand_interpretation)
                #print(type(inst.insn_byte[1]))
                #node_to = inst.insn_byte[1]
                #node_to = int(inst.operand_interpretation.split(',')[-1].split(' ')[-1])
                call_edges.append((node_from, node_to))

    return call_edges


def enum_blocks_edges(function_id, instructions):

    """
    Return a list of basicblock after
    statically parsing given instructions
    """

    basicblocks = list()
    edges = list()

    branches = []
    xrefs = []

    intent = 0
    blocks_tmp = []
    blocks_list = []

    # we need to do that because jump label are relative to the current block index
    for index, inst in enumerate(instructions[:-1]):

        if inst.is_block_terminator:
            start, name = blocks_tmp.pop()
            if inst.name == 'else':
                end = inst.offset - 1
            else:
                end = inst.offset_end
            blocks_list.append((intent, start, end, name))
            intent -= 1
        if inst.is_block_starter:  # in ['block', 'loop']:
            blocks_tmp.append((inst.offset, inst.name))
            intent += 1
        if inst.is_branch:
            branches.append((intent, inst))

    # add function body end
    blocks_list.append((0, 0, instructions[-1].offset_end, 'func'))
    blocks_list = sorted(blocks_list, key=lambda tup: (tup[1], tup[0]))

    for depth, inst in branches:
        labl = list()
        if inst.name == 'br_table':
            labl = [i for i in inst.insn_byte[2:]]
        else:
            labl.append(int(inst.operand_interpretation.split(' ')[-1]))

        for d2 in labl:
            rep = next(((i, s, e, n) for i, s, e, n in blocks_list if (i == (depth - d2) and s < inst.offset and e > inst.offset_end)), None)

            if rep:
                i, start, end, name = rep
                # if we branch to a 'loop' label
                # we go at the entry of the 'loop' block
                if name == 'loop':
                    value = start
                # if we branch to a 'block' label
                # we go at the end of the "block" block
                elif name == 'block' or name == 'func':
                    value = end
                # we don't know
                else:
                    value = None
                inst.xref.append(value)
                xrefs.append(value)

    # assign xref for "if" branch
    # needed because 'if' don't used label
    for index, inst in enumerate(instructions[:-1]):
        if inst.name == 'if':
            g_block = next(iter([b for b in blocks_list if b[1] == inst.offset]), None)
            jump_target = g_block[2] + 1
            inst.xref.append(jump_target)
            xrefs.append(jump_target)
        elif inst.name == 'else':
            g_block = next(iter([b for b in blocks_list if b[1] == inst.offset]), None)
            jump_target = g_block[2] + 1
            inst.xref.append(jump_target)
            xrefs.append(jump_target)

    # enumerate blocks
    new_block = True

    for index, inst in enumerate(instructions):

        # creation of a block
        if new_block:
            block = BasicBlock(inst.offset,
                               inst,
                               name=format_bb_name(function_id, inst.offset))
            new_block = False
        # add current instruction to the basicblock
        block.instructions.append(inst)

        # next instruction is a jump target
        if index < (len(instructions) - 1) and \
                instructions[index + 1].offset in xrefs:
            new_block = True
        # absolute jump - br
        elif inst.is_branch_unconditional:
            new_block = True
        # conditionnal jump - br_if
        elif inst.is_branch_conditional:
            new_block = True
        # is_block_terminator
        # GRAPHICAL OPTIMIZATION: merge end together
        elif index < (len(instructions) - 1) and \
                instructions[index + 1].name in ['else', 'loop']:  # is_block_terminator
            new_block = True
        # last instruction of the bytecode
        elif inst.offset == instructions[-1].offset:
            new_block = True

        if new_block:
            block.end_offset = inst.offset_end
            block.end_instr = inst
            basicblocks.append(block)
            new_block = True

    # enumerate edges
    for index, block in enumerate(basicblocks):
        # get the last instruction
        inst = block.end_instr
        # unconditional jump - br
        if inst.is_branch_unconditional:
            for ref in inst.xref:
                edges.append(Edge(block.name, format_bb_name(function_id, ref), EDGE_UNCONDITIONAL))
        # conditionnal jump - br_if, if
        elif inst.is_branch_conditional:
            if inst.name == 'if':
                edges.append(Edge(block.name,
                             format_bb_name(function_id, inst.offset_end + 1),
                             EDGE_CONDITIONAL_TRUE))
                if_b = next(iter([b for b in blocks_list if b[1] == inst.offset]), None)
                #else_block = blocks_list[blocks_list.index(if_block) + 1]
                jump_target = if_b[2] + 1
                edges.append(Edge(block.name,
                             format_bb_name(function_id, jump_target),
                             EDGE_CONDITIONAL_FALSE))
            else:
                for ref in inst.xref:
                    if ref and ref != inst.offset_end + 1:
                        # create conditionnal true edges
                        edges.append(Edge(block.name,
                                          format_bb_name(function_id, ref),
                                          EDGE_CONDITIONAL_TRUE))
                # create conditionnal false edge
                edges.append(Edge(block.name,
                             format_bb_name(function_id, inst.offset_end + 1),
                             EDGE_CONDITIONAL_FALSE))
        # instruction that end the flow
        elif [i.name for i in block.instructions if i.is_halt]:
            pass
        elif inst.is_halt:
            pass

        # handle the case when you have if and else following
        elif inst.offset != instructions[-1].offset and \
                block.start_instr.name != 'else' and \
                instructions[instructions.index(inst) + 1].name == 'else':

            else_ins = instructions[instructions.index(inst) + 1]
            else_b = next(iter([b for b in blocks_list if b[1] == else_ins.offset]), None)

            edges.append(Edge(block.name, format_bb_name(function_id, else_b[2] + 1), EDGE_FALLTHROUGH))
        # add the last intruction "end" in the last block
        elif inst.offset != instructions[-1].offset:
            # EDGE_FALLTHROUGH
            edges.append(Edge(block.name, format_bb_name(function_id, inst.offset_end + 1), EDGE_FALLTHROUGH))

    # prevent duplicate edges
    edges = list(set(edges))
    return basicblocks, edges


class WasmCFG(CFG):
    """
    """
    def __init__(self, module_bytecode):
        self.module_bytecode = bytecode_to_bytes(module_bytecode)

        self.functions = list()
        self.basicblocks = list()
        self.edges = list()

        self.analyzer = WasmModuleAnalyzer(self.module_bytecode)
        self.run_static_analysis()

    def run_static_analysis(self):
        self.functions = enum_func(self.module_bytecode)

        for idx, func in enumerate(self.functions):
            func.basicblocks, edges = enum_blocks_edges(idx, func.instructions)
            # all bb name are unique so we can create global bb & edge list
            self.basicblocks += func.basicblocks
            self.edges += edges

    def get_function(self, name=None, prefered_name=None):
        if name:
            return [x for x in self.functions if x.name == name][0]
        elif prefered_name:
            return [x for x in self.functions if x.prefered_name == prefered_name][0]
        else:
            raise Exception('name/prefered_name not found, please check again')

    def get_functions_call_edges(self, format_fname=False):

        nodes = list()
        edges = list()

        if not self.analyzer:
            self.analyzer = WasmModuleAnalyzer(self.module_bytecode)
        if not self.functions:
            self.functions = enum_func(self.module_bytecode)

        # create nodes
        for name, param_str, return_str, _ in self.analyzer.func_prototypes:
            if format_fname:
                nodes.append(format_func_name(name, param_str, return_str))
            else:
                nodes.append(name)

        logging.info('nodes: %s', nodes)

        # create edges
        tmp_edges = enum_func_call_edges(self.functions,
                                         len(self.analyzer.imports_func))

        # tmp_edges = [(node_from, node_to), (...), ...]
        for node_from, node_to in tmp_edges:
            # node_from
            name, param, ret, _ = self.analyzer.func_prototypes[node_from]
            if format_fname:
                from_final = format_func_name(name, param, ret)
            else:
                from_final = name
            # node_to
            name, param, ret, _ = self.analyzer.func_prototypes[node_to]
            to_final = format_func_name(name, param, ret)
            if format_fname:
                to_final = format_func_name(name, param, ret)
            else:
                to_final = name
            edges.append(Edge(from_final, to_final, EDGE_CALL))
        logging.info('edges: %s', edges)

        return (nodes, edges)

    def __str__(self):
        line = ("length functions = %d\n" % len(self.functions))
        line += ("length basicblocks = %d\n" % len(self.basicblocks))
        line += ("length edges = %d\n" % len(self.edges))
        return line

    def visualize(self, function=True, simplify=False, ssa=False):
        """Visualize the cfg
        used CFGGraph
        equivalent to:
            graph = CFGGraph(cfg)
            graph.view_functions()
        """
        graph = CFGGraph(self)
        if function:
            graph.view_functions(simplify=simplify, ssa=ssa)
        else:
            graph.view(simplify=simplify, ssa=ssa)

    def visualize_call_flow(self, filename="wasm_call_graph_octopus.gv",
                            format_fname=False):
        """Visualize the cfg call flow graph
        """
        nodes, edges = self.get_functions_call_edges()
        if format_fname:
            nodes_longname, edges = self.get_functions_call_edges(format_fname=True)

        g = Digraph(filename, filename=filename)
        g.attr(rankdir='LR')

        with g.subgraph(name='global') as c:

            export_list = [p[0] for p in self.analyzer.func_prototypes if p[3] == 'export']
            import_list = [p[0] for p in self.analyzer.func_prototypes if p[3] == 'import']
            try:
                indirect_target = [self.analyzer.func_prototypes[index][0] for index in self.analyzer.elements[0].get('elems')]
            except IndexError:
                indirect_target = []
            # create all the graph nodes (function name)
            for idx, node in enumerate(nodes):
                # name graph bubble
                node_name = node
                if format_fname:
                    node_name = nodes_longname[idx]

                # default style value
                fillcolor = "white"
                shape = "ellipse"
                style = "filled"

                if node in import_list:
                    logging.debug('import ' + node)
                    fillcolor = DESIGN_IMPORT.get('fillcolor')
                    shape = DESIGN_IMPORT.get('shape')
                    style = DESIGN_IMPORT.get('style')
                    c.node(node_name, fillcolor=fillcolor, shape=shape, style=style)
                elif node in export_list:
                    logging.debug('export ' + node)
                    fillcolor = DESIGN_EXPORT.get('fillcolor')
                    shape = DESIGN_EXPORT.get('shape')
                    style = DESIGN_EXPORT.get('style')
                    c.node(node_name, fillcolor=fillcolor, shape=shape, style=style)
                elif node in indirect_target:
                    logging.debug('indirect_target ' + node)
                    shape = "hexagon"

                c.node(node_name, fillcolor=fillcolor, shape=shape, style=style)

            # check if multiple same edges
            # in that case, put the number into label
            edges_counter = dict((x, edges.count(x)) for x in set(edges))
            # insert edges on the graph
            for edge, count in edges_counter.items():
                label = None
                if count > 1:
                    label = str(count)
                c.edge(edge.node_from, edge.node_to, label=label)

        g.render(filename, view=True)

    def visualize_instrs_per_funcs(self, show=True, save=True,
                                   out_filename="wasm_func_analytic.png",
                                   fontsize=8):
        """Visualize the instructions repartitions per functions
        """

        import numpy as np
        import matplotlib.pyplot as plt

        final = list()
        datas = list()

        # legend x axis - name functions
        group_names = tuple([func.name for func in self.functions])
        # number of functions
        ind = [x for x, _ in enumerate(self.functions)]

        # list all groups
        all_groups = [v for _, v in _groups.items()]

        # list()
        for func in self.functions:
            data = list()
            group = [i.group for i in func.instructions]
            for g in all_groups:
                data.append(group.count(g))
            datas.append(tuple(data))

        for idx in range(len(all_groups)):
            final.append(tuple([x[idx] for x in datas]))

        # choice color: https://matplotlib.org/users/colormaps.html
        color = iter(plt.cm.gist_rainbow(np.linspace(0, 1, len(all_groups))))
        stack = np.array([0 * len(all_groups)])
        for idx in range(len(all_groups)):
            if idx == 0:
                # first bar
                plt.barh(ind, final[idx], label=all_groups[idx],
                         align='center', color=next(color))
            else:
                plt.barh(ind, final[idx], label=all_groups[idx], left=stack,
                         align='center', color=next(color))

            stack = stack + np.array(final[idx])

        # Rotate x-labels on the x-axis
        plt.yticks(fontsize=fontsize)
        plt.ylim([0, len(self.functions)])
        plt.yticks(ind, group_names)
        plt.ylabel('Functions')
        plt.xlabel('Instructions count')
        plt.legend(loc="lower right")
        plt.title('Instructions count by function and group')

        # save
        if save:
            plt.savefig(out_filename)
        # show
        if show:
            plt.show()



# The MIT License (MIT)
#
# Copyright (c) 2016 Joel Höner <athre0z@zyantific.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# modified code from https://github.com/athre0z/wasm/blob/master/wasm/modtypes.py
# no need of that if PyPI wasm version 1.2 release


"""Provides functions for decoding WASM modules and bytecode."""
from __future__ import print_function, absolute_import, division, unicode_literals

from collections import namedtuple
from wasm.modtypes import ModuleHeader, Section, SEC_UNK, SEC_NAME, NameSubSection
from wasm.opcodes import OPCODE_MAP
from wasm.compat import byte2int


Instruction = namedtuple('Instruction', 'op imm len')
ModuleFragment = namedtuple('ModuleFragment', 'type data')


def decode_bytecode(bytecode):
    """Decodes raw bytecode, yielding `Instruction`s."""
    bytecode_wnd = memoryview(bytecode)
    while bytecode_wnd:
        opcode_id = byte2int(bytecode_wnd[0])
        opcode = OPCODE_MAP[opcode_id]

        if opcode.imm_struct is not None:
            offs, imm, _ = opcode.imm_struct.from_raw(None, bytecode_wnd[1:])
        else:
            imm = None
            offs = 0

        insn_len = 1 + offs
        yield Instruction(opcode, imm, insn_len)
        bytecode_wnd = bytecode_wnd[insn_len:]


def decode_module(module, decode_name_subsections=False):
    """Decodes raw WASM modules, yielding `ModuleFragment`s."""
    module_wnd = memoryview(module)

    # Read & yield module header.
    hdr = ModuleHeader()
    hdr_len, hdr_data, _ = hdr.from_raw(None, module_wnd)
    yield ModuleFragment(hdr, hdr_data)
    module_wnd = module_wnd[hdr_len:]

    # Read & yield sections.
    while module_wnd:
        sec = Section()
        sec_len, sec_data, _ = sec.from_raw(None, module_wnd)

        # If requested, decode name subsections when encountered.
        if (
            decode_name_subsections and
            sec_data.id == SEC_UNK and
            sec_data.name == SEC_NAME
        ):
            sec_wnd = sec_data.payload
            while sec_wnd:
                subsec = NameSubSection()
                subsec_len, subsec_data, _ = subsec.from_raw(None, sec_wnd)
                yield ModuleFragment(subsec, subsec_data)
                sec_wnd = sec_wnd[subsec_len:]
        else:
            yield ModuleFragment(sec, sec_data)

        # fix bug KeyError
        if sec_data.id == SEC_UNK and sec_data.name:
            sec_len -= sec_data.name_len + 1
        module_wnd = module_wnd[sec_len:]



from octopus.arch.evm.disassembler import EvmDisassembler
from octopus.arch.wasm.disassembler import WasmDisassembler


# Etherem smart contract == EVM bytecode
class EthereumDisassembler(object):
    def __new__(cls, bytecode=None, arch='evm'):
        if arch == 'evm':
            return EvmDisassembler(bytecode)
        else:  # Wasm
            return WasmDisassembler(bytecode)



from octopus.arch.wasm.analyzer import WasmModuleAnalyzer


# Eos smart contract == wasm module
class EosAnalyzer(WasmModuleAnalyzer):
    def __init__(self, module_bytecode, analysis=True):
        WasmModuleAnalyzer.__init__(self,
                                    module_bytecode=module_bytecode,
                                    analysis=analysis)



# https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#language-types
LANG_TYPE = {
    # Opcode, Type constructor
    -0x01: 'i32',
    -0x02: 'i64',
    -0x03: 'f32',
    -0x04: 'f64',
    -0x10: 'anyfunc',
    -0x20: 'func',
    -0x40: 'block_type'
}

KIND_TYPE = {
    0: 'function',
    1: 'table',
    2: 'memory',
    3: 'global',
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import sys

from logging import getLogger
logging = getLogger(__name__)


PLATFORM = 'eth-evm'


def error_print(msg):
    print('[X] %s for %s' % (msg, PLATFORM))
    sys.exit()


def main() -> None:
    parser = argparse.ArgumentParser(
        description='Security Analysis tool for WebAssembly module and Blockchain Smart Contracts (BTC/ETH/NEO/EOS)')

    inputs = parser.add_argument_group('Input arguments')
    inputs.add_argument('-r', '--raw',
                        help='hex-encoded bytecode string ("ABcdeF09..." or "0xABcdeF09...")',
                        metavar='BYTECODE')
    inputs.add_argument('-f', '--file',
                        type=argparse.FileType('r'),
                        help='file containing hex-encoded bytecode string',
                        metavar='BYTECODEFILE')
    inputs.add_argument('-a', '--address',
                        help='pull contract from the blockchain',
                        metavar='CONTRACT_ADDRESS')

    features = parser.add_argument_group('Features')
    features.add_argument('-e', '--explore',
                          action='store_true',
                          help='client to retrieve information from blockchains')
    features.add_argument('-d', '--disassemble',
                          action='store_true',
                          help='print text disassembly ')
    features.add_argument('-g', '--cfg',
                          action='store_true',
                          help='generate the control flow graph (CFG)')
    features.add_argument('-c', '--call',
                          action='store_true',
                          help='generate the call flow graph')
    features.add_argument('-s', '--ssa',
                          action='store_true',
                          help='generate the CFG with SSA representation')

    graph = parser.add_argument_group('Graph options')
    graph.add_argument('--onlystatic', action='store_true',
                       help='generate the CFG without stack emulation (fastest but less accurate)')
    graph.add_argument('--simplify', action='store_true',
                       help='generate a simplify CFG')
    graph.add_argument('--functions', action='store_true',
                       help='create subgraph for each function')
    graph.add_argument('--onlyfunction', type=str,
                       nargs="*",
                       default=[],
                       help='only generate the CFG for this list of function name')
    #graph.add_argument('--visualize',
    #                   help='direcly open the CFG file')
    #graph.add_argument('--format',
    #                   choices=['pdf', 'png', 'dot'],
    #                   default='pdf',
    #                   help='direcly open the CFG file')

    explorer = parser.add_argument_group('Explorer options')
    explorer.add_argument('--code',
                          help='get contract code',
                          metavar='CONTRACT_ADDRESS')
    explorer.add_argument('--tx',
                          help='get transaction information')
    #explorer.add_argument('--txdecode', action='store_true',
    #                      help='return important transaction information')
    explorer.add_argument('--blockid', type=int,
                          help='get block information using given block number',
                          metavar='BLOCK_NUMBER')
    explorer.add_argument('--blockhash',
                          help='get block information using given block hash',
                          metavar='BLOCK_HASH')
    explorer.add_argument('--infura',
                          help='Infura network choice',
                          choices=['mainnet', 'ropsten', 'infuranet', 'kovan', 'rinkeby'],
                          default='mainnet')
    explorer.add_argument('--rpc', help='custom RPC settings',
                          metavar='HOST:PORT')
    explorer.add_argument('--rpctls', action='store_false',
                          help='RPC connection over TLS')

    args = parser.parse_args()

    octo_bytecode = None
    octo_explorer = None
    octo_disasm = None
    octo_cfg = None

    # Explorer
    if args.explore or args.address:
        # user choose some explorer options
        if args.rpc:
            from octopus.platforms.ETH.explorer import EthereumExplorerRPC
            # parsing RPC HOST & PORT
            host, port = args.rpc.split(':')
            octo_explorer = EthereumExplorerRPC(host=host, port=port, tls=args.rpctls)
        else:
            from octopus.platforms.ETH.explorer import EthereumInfuraExplorer
            from octopus.platforms.ETH.explorer import (INFURA_MAINNET,
                                                        INFURA_ROPSTEN,
                                                        INFURA_INFURANET,
                                                        INFURA_KOVAN,
                                                        INFURA_RINKEBY)
            if args.infura == 'mainnet':
                network = INFURA_MAINNET
            if args.infura == 'ropsten':
                network = INFURA_ROPSTEN
            if args.infura == 'infuranet':
                network = INFURA_INFURANET
            if args.infura == 'kovan':
                network = INFURA_KOVAN
            if args.infura == 'rinkeby':
                network = INFURA_RINKEBY

            octo_explorer = EthereumInfuraExplorer(network=network)

        # process explorer related commands
        if args.code:
            print(octo_explorer.eth_getCode(args.code))
        elif args.tx:
            print(octo_explorer.get_transaction(args.tx))
        # elif args.txdecode:
        #    pass
        elif args.blockid:
            print(octo_explorer.get_block_by_number(args.blockid))
        elif args.blockhash:
            print(octo_explorer.get_block_by_hash(args.blockhash))

    # process input code
    if args.raw:
        octo_bytecode = args.raw
    elif args.file:
        octo_bytecode = ''.join([l.strip() for l in args.file if len(l.strip()) > 0])
    elif args.address:
        octo_bytecode = octo_explorer.eth_getCode(args.address)

    # Disassembly
    if args.disassemble:
        from octopus.platforms.ETH.disassembler import EthereumDisassembler

        # TODO add other r_format support
        octo_disasm = EthereumDisassembler()
        print(octo_disasm.disassemble(octo_bytecode, r_format='text'))

    # Control Flow Analysis
    elif args.cfg or args.ssa:
        from octopus.platforms.ETH.cfg import EthereumCFG
        from octopus.analysis.graph import CFGGraph

        if args.onlystatic and not args.ssa:
            octo_cfg = EthereumCFG(octo_bytecode, evm_analysis='static')
        else:
            octo_cfg = EthereumCFG(octo_bytecode)

        octo_graph = CFGGraph(octo_cfg)

        if args.functions or args.onlyfunction:
            octo_graph.view_functions(only_func_name=args.onlyfunction,
                                      simplify=args.simplify,
                                      ssa=args.ssa)
        else:
            octo_graph.view(simplify=args.simplify, ssa=args.ssa)

    # Call Flow Analysis
    elif args.call:
        error_print('Call Flow Analysis not yet supported')
    elif args.ssa:
        # already done
        pass
    elif args.explore:
        # already done
        pass
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



from octopus.core.contract import Contract


class ContractAdressEmptyException(Exception):
    """Exception raised when address is None"""
    pass


class EthereumContract(Contract):

    @property
    def _address_defined(self):
        if self.address is None:
            raise ContractAdressEmptyException()
        return True

    def get_online_bytecode(self, explorer):
        if self._address_defined:
            self.bytecode = explorer.eth_getCode(self.address)

    def get_online_abi(self, explorer):
        raise NotImplementedError

    def get_online_source(self, explorer):
        raise NotImplementedError

    def get_online_info(self, explorer):
        # get the balance
        if self._address_defined:
            # self.name
            self.balance = explorer.eth_getBalance(self.address)
            self.bytecode = explorer.eth_getCode(self.address)
            # self.abi
            # self.source_code



from octopus.engine.explorer import Explorer

# Inspired by https://github.com/ellmetha/neojsonrpc

NEO_HOST = 'seed2.neo.org'

MAINNET_HTTP_RPC_PORT = 10332
MAINNET_HTTPS_RPC_PORT = 10331
TESTNET_HTTP_RPC_PORT = 20332
TESTNET_HTTPS_RPC_PORT = 20331


class NeoExplorerRPC(Explorer):
    '''
    Neo JSON-RPC client class
    '''

    def __init__(self, host=NEO_HOST, port=MAINNET_HTTP_RPC_PORT, tls=False, max_retries=3):
        Explorer.__init__(self, host=host, port=port, tls=tls, max_retries=max_retries)

        # Initializes an "ID counter" that'll be used to forge each request to the JSON-RPC
        # endpoint. The "id" parameter is "required" in order to help clients sort responses out.
        # In the case of the current client, we'll just ensure that this value gets incremented
        # after each request made to the JSON-RPC endpoint.
        self._id_counter = 0

    def call(self, method, params=None, jsonrpc='2.0', _id=None):

        # Determines which 'id' value to use and increment the counter associated with the current
        # client instance if applicable.
        rid = _id or self._id_counter
        if _id is None:
            self._id_counter += 1

        return super().call(method, params, jsonrpc, rid)

    #######################
    # HIGHT-LEVEL METHODS #
    #######################

    def get_transaction(self, transaction_id, verbosity=True):
        """ Return transaction informations

        .. seealso::
            :method:`getrawtransaction`
        """
        return self.getrawtransaction(transaction_id, verbosity)

    def get_block_by_number(self, block_number):
        """ Return block information using given block number

        .. seealso::
            :method:`getblock`
        """
        return self.getblock(block_number)

    def get_block_by_hash(self, block_hash):
        """ Return block information using given block hash

        .. seealso::
            :method:`getblock`
        """
        return self.getblock(block_hash)

    ####################
    # JSON-RPC METHODS #
    ####################

    # ressources :
    # http://docs.neo.org/en-us/node/api.html
    # https://neotracker.io/

    def dumpprivkey(self, address):
        """ Returns the private key of the standard address.

        :param address: a 34-bit length address (eg. AJBENSwajTzQtwyJFkiJSv7MAaaMc7DsRz)
        :type address: str
        :return: Returns the private key of the standard address.
        :rtype: str

        .. note::
            You need to open the wallet in the NEO-CLI node before you execute this command.
        .. seealso::
            http://docs.neo.org/en-us/node/api/dumpprivkey.html
        .. todo::
            SEEM TO WORK
        """
        return self.call('dumpprivkey', [address])

    def getaccountstate(self, address):
        """ Returns the account state information associated with a specific address.

        :param address: a 34-bit length address (eg. AJBENSwajTzQtwyJFkiJSv7MAaaMc7DsRz)
        :type address: str
        :return: dictionary containing the account state information
        :rtype: dict

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getaccountstate("AJBENSwajTzQtwyJFkiJSv7MAaaMc7DsRz")
        {'balances': [{'asset': '0xc56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
            'value': '60'},
        {'asset': '0x602c79718b16e442de58778e148d0b1084e3b2dffd5de6b7b16cee7969282de7',
            'value': '0.29359449'}],
        'frozen': False,
        'script_hash': '0xb6be270ca7cf0affe93de076fc3ace8662027f61',
        'version': 0,
        'votes': []}

        .. seealso::
            http://docs.neo.org/en-us/node/api/getaccountstate.html
        .. todo::
            TESTED

        """
        return self.call('getaccountstate', [address])

    def getapplicationlog(self, txid):
        """ Returns the contract log based on the specified txid.

        :param txid: Transaction ID
        :type txid: str
        :return: dictionary containing the contract logs
        :rtype: dict

        .. note::
            You need to run the command dotnet neo-cli.dll --log to enable logging before invoking this method. The complete contract logs are stored under the ApplicationLogs directory.
        .. seealso::
            http://docs.neo.org/en-us/node/api/getapplicationlog.html
        .. todo::
            NOT TESTED
        """
        return self.call('getapplicationlog', [txid])

    def getassetstate(self, asset_id):
        """ Returns the asset information associated with a specific asset ID.

        :param asset_id:
            an asset identifier (the transaction ID of the RegistTransaction when the asset is
            registered)
        :type asset_id: str
        :return: dictionary containing the asset state information
        :rtype: dict

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getassetstate("c56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b")
        {'admin': 'Abf2qMs1pzQb8kYk9RuxtUb9jtRKJVuBJt',
         'amount': '100000000',
         'available': '100000000',
         'expiration': 4000000,
         'frozen': False,
         'id': '0xc56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
         'issuer': 'Abf2qMs1pzQb8kYk9RuxtUb9jtRKJVuBJt',
         'name': [{'lang': 'zh-CN', 'name': '小蚁股'},
         {'lang': 'en', 'name': 'AntShare'}],
         'owner': '00',
         'precision': 0,
         'type': 'GoverningToken',
         'version': 0}

        .. seealso::
            http://docs.neo.org/en-us/node/api/getassetstate.html
        .. todo::
            TESTED
        """
        return self.call('getassetstate', [asset_id])

    def getbalance(self, asset_id):
        """ Returns the balance of the corresponding asset in the wallet, based on the specified asset number.

        :param asset_id:
            an asset identifier (the transaction ID of the RegistTransaction when the asset is
            registered)
        :type asset_id: str
        :return: dictionary containing the actual balance and the exact amount of the asset in the wallet
        :rtype: dict

        .. note::
            You need to open the wallet in the NEO-CLI node before you execute this command.
        .. seealso::
            http://docs.neo.org/en-us/node/api/getbalance.html
        .. todo::
            NOT TESTED
        """
        return self.call('getbalance', [asset_id])

    def getbestblockhash(self):
        """ Returns the hash of the tallest block in the main chain.
        :return: hash of the tallest block in the chain
        :rtype: str

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getbestblockhash()
        '0x881da7e14680cd4a020aa503dc602a92a411ad3184b6b14789074c775fbe5b7b'

        .. seealso::
            http://docs.neo.org/en-us/node/api/getbestblockhash.html
        .. todo::
            TESTED
        """
        return self.call('getbestblockhash')

    def getblock(self, block_id, verbose=True):
        """ Returns the block information associated with a specific hash value or block index.

        :param block_hash: a block hash value or a block index (block height)
        :param verbose:
            a boolean indicating whether the detailed block information should be returned in JSON
            format (otherwise the block information is returned as an hexadecimal string by the
            JSON-RPC endpoint)
        :type block_hash: str or int
        :type verbose: bool
        :return:
            dictionary containing the block information (or an hexadecimal string if verbose is set
            to False)
        :rtype: dict or str

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getblock(1917115)
        {'confirmations': 1,
         'hash': '0x881da7e14680cd4a020aa503dc602a92a411ad3184b6b14789074c775fbe5b7b',
         'index': 1917115,
         .....
        >>> explorer.getblock('0x881da7e14680cd4a020aa503dc602a92a411ad3184b6b14789074c775fbe5b7b')
        {'confirmations': 1,
         'hash': '0x881da7e14680cd4a020aa503dc602a92a411ad3184b6b14789074c775fbe5b7b',
         'index': 1917115,

        .. seealso::
            http://docs.neo.org/en-us/node/api/getblock.html
            http://docs.neo.org/en-us/node/api/getblock2.html
        .. todo::
            TESTED
        """
        return self.call('getblock', [block_id, int(verbose)])

    def getblockcount(self):
        """ Returns the number of blocks in the chain.
        :return: number of blocks in the chain
        :rtype: int

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getblockcount()
        1917116

        .. seealso::
            http://docs.neo.org/en-us/node/api/getblockcount.html
        .. todo::
            TESTED
        """
        return self.call('getblockcount')

    def getblockhash(self, block_index):
        """ Returns the hash value associated with a specific block index.
        :param block_index: a block index (block height)
        :type block_index: int
        :return: hash of the block associated with the considered index
        :rtype: str

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getblockhash(1917115)
        '0x881da7e14680cd4a020aa503dc602a92a411ad3184b6b14789074c775fbe5b7b'

        .. seealso::
            http://docs.neo.org/en-us/node/api/getblockhash.html
        .. todo::
            TESTED
        """
        return self.call('getblockhash', [block_index])

    def getblocksysfee(self, block_index):
        """ Returns the system fees associated with a specific block index.
        :param block_index: a block index (block height)
        :type block_index: int
        :return: system fees of the block, expressed in NeoGas units
        :rtype: str

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getblocksysfee(1917115)
        '206594'

        .. seealso::
            http://docs.neo.org/en-us/node/api/getblocksysfee.html
        .. todo::
            TESTED
        """
        return self.call('getblocksysfee', [block_index])

    def getconnectioncount(self):
        """ Returns the current number of connections for the considered node.
        :return: number of connections for the node
        :rtype: int

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getconnectioncount()
        57

        .. seealso::
            http://docs.neo.org/en-us/node/api/getconnectioncount.html
        .. todo::
            TESTED
        """
        return self.call('getconnectioncount')

    def getcontractstate(self, script_hash):
        """ Returns the contract information associated with a specific script hash.
        :param script_hash: contract script hash
        :type script_hash: str
        :return: dictionary containing the contract information
        :rtype: dict

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getcontractstate("d3cce84d0800172d09c88ccad61130611bd047a4")
        {'author': 'Erik Zhang',
         'code_version': '2.0',
         'description': 'Lock 2.0',
         'email': 'erik@neo.org',
         'hash': '0xd3cce84d0800172d09c88ccad61130611bd047a4',
         'name': 'Lock',
         'parameters': ['Integer', 'PublicKey', 'Signature'],
         'returntype': 'Boolean',
         'script': '56c56b6c766b00527ac46c766b51527ac46c766b52527ac4616168184e656f2e426c6f636b636861696e2e4765744865696768746168184e656f2e426c6f636b636861696e2e4765744865616465726c766b53527ac46c766b00c36c766b53c36168174e656f2e4865616465722e47657454696d657374616d70a06c766b54527ac46c766b54c3640e00006c766b55527ac4621a006c766b51c36c766b52c3617cac6c766b55527ac46203006c766b55c3616c7566',
         'storage': False,
         'version': 0}

        .. seealso::
            http://docs.neo.org/en-us/node/api/getcontractstate.html
        .. todo::
            TESTED
        """
        return self.call('getcontractstate', [script_hash])

    def getnewaddress(self):
        """ Creates a new address and return it.
        :return: the newly created address.
        :rtype: str

        .. note::
            You need to open the wallet in the NEO-CLI node before you execute this command.
        .. seealso::
            http://docs.neo.org/en-us/node/api/getnewaddress.html
        .. todo::
            NOT TESTED
        """
        return self.call('getnewaddress')

    def getrawmempool(self):
        """ Returns a list of unconfirmed transactions in memory associated with the node.
        :return: list of unconfirmed transaction hashes
        :rtype: list

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.getrawmempool()
        ['0xb7ed3825dfb23db0198df1d00b7250a867b1b7457b0f5ec9c0f9393ad1939729',
         '0x4393c79340eb9e11fe6a179984b6989195044c79b92dee16a8a1090189b8ef4d',
         '0xa705e48ab47817409028d89bf5727399333cdefe60d879ac5b70df38f3096436',
         ...

        .. seealso::
            http://docs.neo.org/en-us/node/api/getrawmempool.html
        .. todo::
            TESTED
        """
        return self.call('getrawmempool')

    def getrawtransaction(self, tx_hash, verbose=True):
        """ Returns detailed information associated with a specific transaction hash.

        :param tx_hash: transaction hash
        :param verbose:
            a boolean indicating whether the detailed transaction information should be returned in
            JSON format (otherwise the transaction information is returned as an hexadecimal string
            by the JSON-RPC endpoint)
        :type tx_hash: str
        :type verbose: bool
        :return:
            dictionary containing the transaction information (or an hexadecimal string if verbose
            is set to False)
        :rtype: dict or str

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> tx_id = "f4250dab094c38d8265acc15c366dc508d2e14bf5699e12d9df26577ed74d657"
        >>> explorer.getrawtransaction(tx_id, False)
        '80000001195876cb34364dc38b730077156c6bc3a7fc570044a66fbfeeea56f71327e8ab000[...]dac'
        >>> explorer.getrawtransaction(tx_id)
        {'attributes': [],
         'blockhash': '0x9c814276156d33f5dbd4e1bd4e279bb4da4ca73ea7b7f9f0833231854648a72c',
         'blocktime': 1496719422,
         'confirmations': 925250,
         'net_fee': '0',
         'scripts': [{'invocation': '40915467ecd359684b2dc358024ca750609591aa731a0b309c7fb3cab5cd0836ad3992aa0a24da431f43b68883ea5651d548feb6bd3c8e16376e6e426f91f84c58',
           'verification': '2103322f35c7819267e721335948d385fae5be66e7ba8c748ac15467dcca0693692dac'}],
         'size': 262,
         'sys_fee': '0',
         'txid': '0xf4250dab094c38d8265acc15c366dc508d2e14bf5699e12d9df26577ed74d657',
         'type': 'ContractTransaction',
         'version': 0,
         'vin': [{'txid': '0xabe82713f756eaeebf6fa6440057fca7c36b6c157700738bc34d3634cb765819',
           'vout': 0}],
         'vout': [{'address': 'AHCNSDkh2Xs66SzmyKGdoDKY752uyeXDrt',
           'asset': '0xc56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
           'n': 0,
           'value': '2950'},
          {'address': 'ALDCagdWUVV4wYoEzCcJ4dtHqtWhsNEEaR',
           'asset': '0xc56f33fc6ecfcd0c225c4ab356fee59390af8560be0e930faebe74a6daff7c9b',
           'n': 1,
           'value': '4050'}]}

        .. seealso::
            http://docs.neo.org/en-us/node/api/getrawtransaction.html
        .. todo::
            TESTED
        """
        return self.call('getrawtransaction', [tx_hash, int(verbose)])

    def getstorage(self, script_hash, hexkey):
        """ Returns the value stored in the storage of a contract script hash for a given key.

        :param script_hash: contract script hash
        :param hexkey: key to look up in the storage (in hex string)
        :type script_hash: str
        :type key: str
        :return: the stored value (in hex string)
        :rtype: str

        .. seealso::
            http://docs.neo.org/en-us/node/api/getstorage.html
        .. todo::
            SEEM TO WORK
        """
        return self.call('getstorage', [script_hash, hexkey])

    def gettxout(self, tx_hash, index):
        """ Returns the transaction output information corresponding to a hash and index.

        :param tx_hash: transaction hash
        :param index:
            index of the transaction output to be obtained in the transaction (starts from 0)
        :type tx_hash: str
        :type index: int
        :return: dictionary containing the transaction output
            If the transaction output is already spent, the result value will be null.
        :rtype: dict

        .. seealso::
            http://docs.neo.org/en-us/node/api/gettxout.html
        .. todo::
            SEEM TO WORK
        """
        return self.call('gettxout', [tx_hash, index])

    def getpeers(self):
        """ Returns a list of nodes that the node is currently connected/disconnected from.
        :return: dictionary containing the nodes the current node is connected/disconnected from
        :rtype: dict

        .. seealso::
            http://docs.neo.org/en-us/node/api/getpeers.html
        .. todo::
            NOT TESTED
        """
        return self.call('getpeers', [])

    def getversion(self):
        """ Returns version information about the current node.
        :return: dictionary containing version information about the current node
        :rtype: dict

        .. seealso::
            http://docs.neo.org/en-us/node/api/getversion.html
        .. todo::
            NOT TESTED
        """
        return self.call('getversion')

    def invoke(self, script_hash, params):
        """ Invokes a contract with given parameters and returns the result.
        It should be noted that the name of the function invoked in the contract should be part of
        paramaters.

        :param script_hash: contract script hash
        :param params: list of parameters to be passed in to the smart contract
        :type script_hash: str
        :type params: list
        :return: result of the invocation
        :rtype: dictionary

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.invoke("dc675afc61a7c0f7b3d2682bf6e1d8ed865a0e5f",
            [{"type": "String","value": "name"},{"type":"Boolean","value": False}])
        {'gas_consumed': '0.01',
         'stack': [{'type': 'ByteArray', 'value': ''},
          {'type': 'ByteArray', 'value': '6e616d65'}],
         'state': 'FAULT, BREAK'}

        .. note::
            This method is to test your VM script as if they were ran on the blockchain at that point in time. This RPC call does not affect the blockchain in any way.
        .. seealso::
            http://docs.neo.org/en-us/node/api/invoke.html
        .. todo::
            SEEM TO WORK
        """
        return self.call('invoke', [script_hash, params])

    def invokefunction(self, script_hash, operation, params):
        """ Invokes a contract's function with given parameters and returns the result.

        :param script_hash: contract script hash
        :param operation: name of the operation to invoke
        :param params: list of paramaters to be passed in to the smart contract
        :type script_hash: str
        :type operation: str
        :type params: list
        :return: result of the invocation
        :rtype: dictionary

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> script_hash = "ecc6b20d3ccac1ee9ef109af5a7cdb85706b1df9"
        >>> explorer.invokefunction(script_hash, "balanceOf", [{"type": "Hash160", "value": "bfc469dd56932409677278f6b7422f3e1f34481d"}])
        {'gas_consumed': '0.338',
         'stack': [{'type': 'ByteArray', 'value': '00ab510d'}],
         'state': 'HALT, BREAK'}

        .. note::
            This method is to test your VM script as if they were ran on the blockchain at that point in time. This RPC call does not affect the blockchain in any way.
        .. seealso::
            http://docs.neo.org/en-us/node/api/invokefunction.html
        .. todo::
            TESTED
        """
        return self.call('invokefunction', [script_hash, operation, params])

    def invokescript(self, script):
        """ Invokes a script on the VM and returns the result.
        :param script: script runnable by the VM
        :type script: str
        :return: result of the invocation
        :rtype: dictionary

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.invokescript("00046e616d656711c4d1f4fba619f2628870d36e3a9773e874705b")
        {'gas_consumed': '0.01',
         'stack': [{'type': 'ByteArray', 'value': ''},
          {'type': 'ByteArray', 'value': '6e616d65'}],
         'state': 'FAULT, BREAK'}

        .. note::
            This method is to test your VM script as if they were ran on the blockchain at that point in time. This RPC call does not affect the blockchain in any way.
        .. seealso::
            http://docs.neo.org/en-us/node/api/invokescript.html
        .. todo::
            TESTED
        """
        return self.call('invokescript', [script])

    def listaddress(self):
        """ Lists all the addresses in the current wallet.
        :return: dictionary containing addresses in the wallet and indicates whether it is a watch only address.
        :rtype: dict

        .. note::
            You need to open the wallet in the NEO-CLI node before invoking this method.
        .. seealso::
            http://docs.neo.org/en-us/node/api/listaddress.html
        .. todo::
            NOT TESTED
        """
        return self.call('listaddress')

    def sendrawtransaction(self, hextx):
        """ Broadcasts a transaction over the NEO network and returns the result.
        :param hextx: hexadecimal string that has been serialized
        :type hextx: str
        :return: result of the transaction
        :rtype: bool

        .. seealso::
            http://docs.neo.org/en-us/node/api/sendrawtransaction.html
        .. todo::
            NOT TESTED
        """
        return self.call('sendrawtransaction', [hextx])

    def sendtoaddress(self, asset_id, address, value, fee=0):
        """ Transfers to the specified address.

        :param asset_id:
            an asset identifier (the transaction ID of the RegistTransaction when the asset is
            registered)
        :type asset_id: str
        :param address: a 34-bit length address (eg. AJBENSwajTzQtwyJFkiJSv7MAaaMc7DsRz)
        :type address: str
        :param value: amount to transfer
        :type value: int
        :param fee: Fee, default value is 0 (optional parameter)
        :type fee: int
        :return:
            dictionary containing the asset state information
            If the signature is incomplete, it will return the transaction to be signed.
            If the balance is insufficient, it will return an error message.
        :rtype: dict or str

        .. note::
            You need to open the wallet in the Neo-CLI node before you execute this command.
        .. seealso::
            http://docs.neo.org/en-us/node/api/sendtoaddress.html
        .. todo::
            NOT TESTED
        """
        return self.call('sendtoaddress', [asset_id, address, value, fee])

    def sendmany(self, outputs_array, fee=0, change_address=None):
        """ Bulk transfer order, and you can specify a change address.

        :param outputs_array: Array, the data structure of each element in the array is as follows:
            {"asset": <asset>,"value": <value>,"address": <address>}
            :param asset:
                an asset identifier (the transaction ID of the RegistTransaction when the asset is
                registered)
            :type asset: str
            :param value: Transfer amount
            :type value: int
            :param address: a 34-bit length destination address (eg. AJBENSwajTzQtwyJFkiJSv7MAaaMc7DsRz)
            :type address: str
        :type outputs_array: list
        :param fee: Handling fee, optional parameter, default is 0.
        :type fee: int
        :param change_address: Change address, optional parameter, default is the first standard address in the wallet.
        :type change_address: str
        :return:
            Returns the transaction details as above to indicate that the transaction was sent successfully or the transaction failed.
            If the JSON format is incorrect, a Parse error is returned.
            If the signature is incomplete, a pending transaction is returned.
            If the balance is insufficient, an error message is returned.
        :rtype: dict or str

        .. note::
            You need to open the wallet in the Neo-CLI node before you execute this command.
        .. seealso::
            http://docs.neo.org/en-us/node/api/sendmany.html
        .. todo::
            NOT TESTED
        """
        return self.call('sendmany')

    def submitblock(self, hex_string):
        """ Submit new blocks

        No documentation about this RPC methods (but described as API)

        .. note::
            Needs to be a consensus node
        .. seealso::
            https://github.com/neo-project/docs/blob/master/en-us/node/api.md
        .. todo::
            NOT TESTED
        """
        return self.call('submitblock', [hex_string])

    def validateaddress(self, address):
        """ Validates if the considered string is a valid NEO address.

        :param address: string containing a potential NEO address
        :type address: str
        :return: dictionary containing the result of the verification
        :rtype: dictionary

        :Example:

        >>> explorer = NeoExplorerRPC()
        >>> explorer.validateaddress("AJBENSwajTzQtwyJFkiJSv7MAaaMc7DsRz")
        {'address': 'AJBENSwajTzQtwyJFkiJSv7MAaaMc7DsRz', 'isvalid': True}

        .. seealso::
            http://docs.neo.org/en-us/node/api/validateaddress.html

        """
        return self.call('validateaddress', [address])



from octopus.core.instruction import Instruction


class NeoInstruction(Instruction):
    """NEO Instruction

    TODO

    """
    def __init__(self, opcode, name,
                 operand_size, pops, pushes, fee,
                 description, operand=None,
                 operand_interpretation=None, offset=0, xref=None):
        """ TODO """
        super().__init__(opcode=opcode, name=name,
                         operand_size=operand_size, pops=pops, pushes=pushes,
                         fee=fee, description=description, operand=operand,
                         operand_interpretation=operand_interpretation,
                         offset=offset)

    @property
    def has_length_operand(self):
        """ True if the instruction uses an immediate operand to define the lenght of operand """
        return self.name in ['SYSCALL']

    # overwrite this function to take care of has_length_operand
    @property
    def size(self):
        """ Size of the encoded instruction """
        if self.has_length_operand:
            # add size of lengh operand
            return self.operand_size + 2
        return self.operand_size + 1

    @property
    def group(self):
        """ Instruction classification per group """
        classes = {0x00: 'Constant',
                   0x61: 'Flow control',
                   0x6A: 'Stack',
                   0x7E: 'Splice',
                   0x83: 'Bitwise logic',
                   0x8B: 'Arithmetic',
                   0xA6: 'Crypto',
                   0xC0: 'Array',
                   0xF0: 'Exceptions'}

        last_class = classes.get(0)
        for k, v in classes.items():
            if self.opcode >= k:
                last_class = v
            else:
                return last_class
        return last_class

    @property
    def is_constant(self):
        return self.group == 'Constant'

    @property
    def is_flow_control(self):
        return self.group == 'Flow control'

    @property
    def is_stack(self):
        return self.group == 'Stack'

    @property
    def is_splice(self):
        return self.group == 'Splice'

    @property
    def is_bitwise_logic(self):
        return self.group == 'Bitwise logic'

    @property
    def is_arithmetic(self):
        return self.group == 'Arithmetic'

    @property
    def is_crypto(self):
        return self.group == 'Crypto'

    @property
    def is_array(self):
        return self.group == 'Array'

    @property
    def is_exceptions(self):
        return self.group == 'Exceptions'

    @property
    def is_branch_conditional(self):
        """ Return list if the instruction is a jump """
        return self.name in ['JMPIF', 'JMPIFNOT']

    @property
    def is_branch_unconditional(self):
        """ Return list if the instruction is a jump """
        return self.name in ['JMP']

    @property
    def is_branch(self):
        """ True if the instruction is a jump """
        return self.is_branch_conditional or self.is_branch_unconditional

    @property
    def is_halt(self):
        """ Return list if the instruction is a basic block terminator """
        return (self.name in ['RET'] or self.is_exceptions)

    @property
    def is_terminator(self):
        """ True if the instruction is a basic block terminator """
        return self.is_branch or self.is_halt

    @property
    def have_xref(self):
        """ True if the instruction operand is a reference to an other instruction """
        return self.is_branch or (self.name in ['CALL'])



# extract from:
# * https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md
# * https://webassembly.github.io/spec/core/binary/instructions.html
# * https://github.com/athre0z/wasm/blob/master/wasm/opcodes.py

from wasm.immtypes import *

_groups = {0x00: 'Control',
           0x1A: 'Parametric',
           0x20: 'Variable',
           0x28: 'Memory',
           0x41: 'Constant',
           0x45: 'Logical_i32',
           0x50: 'Logical_i64',
           0x5b: 'Logical_f32',
           0x61: 'Logical_f64',
           0x67: 'Arithmetic_i32',
           0x71: 'Bitwise_i32',
           0x79: 'Arithmetic_i64',
           0x83: 'Bitwise_i64',
           0x8b: 'Arithmetic_f32',
           0x99: 'Arithmetic_f64',
           0xa7: 'Conversion'}

_table = {
    # opcode:(mnemonic/name, imm_struct, pops, pushes, description)
    0x00: ('unreachable', None, 0, 0, 'trap immediately'),
    0x01: ('nop', None, 0, 0, 'no operation'),
    0x02: ('block', BlockImm(), 0, 0, 'begin a sequence of expressions'),
    0x03: ('loop', BlockImm(), 0, 0, 'begin a block which can also form control flow loops'),
    0x04: ('if', BlockImm(), 1, 0, 'begin if expression'),
    0x05: ('else', None, 0, 0, 'begin else expression of if'),
    0x0b: ('end', None, 0, 0, 'end a block, loop, or if'),
    0x0c: ('br', BranchImm(), 0, 0, 'break that targets an outer nested block'),
    0x0d: ('br_if', BranchImm(), 1, 0, 'conditional break that targets an outer nested block'),
    0x0e: ('br_table', BranchTableImm(), 1, 0, 'branch table control flow construct'),
    0x0f: ('return', None, 1, 0, 'return zero or one value from this function'),
    0x10: ('call', CallImm(), 0, 0, 'call a function by its index'),
    0x11: ('call_indirect', CallIndirectImm(), 1, 0, 'call a function indirect with an expected signature'),

    0x1a: ('drop', None, 1, 0, 'ignore value'),
    0x1b: ('select', None, 3, 1, 'select one of two values based on condition'),

    0x20: ('get_local', LocalVarXsImm(), 0, 1, 'read a local variable or parameter'),
    0x21: ('set_local', LocalVarXsImm(), 1, 0, 'write a local variable or parameter'),
    0x22: ('tee_local', LocalVarXsImm(), 0, 1, 'write a local variable or parameter and return the same value'),
    0x23: ('get_global', GlobalVarXsImm(), 0, 1, 'read a global variable'),
    0x24: ('set_global', GlobalVarXsImm(), 1, 0, 'write a global variable'),

    0x28: ('i32.load', MemoryImm(), 1, 1, 'load from memory'),
    0x29: ('i64.load', MemoryImm(), 1, 1, 'load from memory'),
    0x2a: ('f32.load', MemoryImm(), 1, 1, 'load from memory'),
    0x2b: ('f64.load', MemoryImm(), 1, 1, 'load from memory'),
    0x2c: ('i32.load8_s', MemoryImm(), 1, 1, 'load from memory'),
    0x2d: ('i32.load8_u', MemoryImm(), 1, 1, 'load from memory'),
    0x2e: ('i32.load16_s', MemoryImm(), 1, 1, 'load from memory'),
    0x2f: ('i32.load16_u', MemoryImm(), 1, 1, 'load from memory'),
    0x30: ('i64.load8_s', MemoryImm(), 1, 1, 'load from memory'),
    0x31: ('i64.load8_u', MemoryImm(), 1, 1, 'load from memory'),
    0x32: ('i64.load16_s', MemoryImm(), 1, 1, 'load from memory'),
    0x33: ('i64.load16_u', MemoryImm(), 1, 1, 'load from memory'),
    0x34: ('i64.load32_s', MemoryImm(), 1, 1, 'load from memory'),
    0x35: ('i64.load32_u', MemoryImm(), 1, 1, 'load from memory'),
    0x36: ('i32.store', MemoryImm(), 2, 0, 'store to memory'),
    0x37: ('i64.store', MemoryImm(), 2, 0, 'store to memory'),
    0x38: ('f32.store', MemoryImm(), 2, 0, 'store to memory'),
    0x39: ('f64.store', MemoryImm(), 2, 0, 'store to memory'),
    0x3a: ('i32.store8', MemoryImm(), 2, 0, 'store to memory'),
    0x3b: ('i32.store16', MemoryImm(), 2, 0, 'store to memory'),
    0x3c: ('i64.store8', MemoryImm(), 2, 0, 'store to memory'),
    0x3d: ('i64.store16', MemoryImm(), 2, 0, 'store to memory'),
    0x3e: ('i64.store32', MemoryImm(), 2, 0, 'store to memory'),
    0x3f: ('current_memory', CurGrowMemImm(), 0, 1, 'query the size of memory'),
    0x40: ('grow_memory', CurGrowMemImm(), 0, 0, 'grow the size of memory'),

    0x41: ('i32.const', I32ConstImm(), 0, 1, 'a constant value interpreted as i32'),
    0x42: ('i64.const', I64ConstImm(), 0, 1, 'a constant value interpreted as i64'),
    0x43: ('f32.const', F32ConstImm(), 0, 1, 'a constant value interpreted as f32'),
    0x44: ('f64.const', F64ConstImm(), 0, 1, 'a constant value interpreted as f64'),

    0x45: ('i32.eqz', None, 1, 1, 'compare equal to zero (return 1 if operand is zero, 0 otherwise)'),
    0x46: ('i32.eq', None, 2, 1, 'sign-agnostic compare equal'),
    0x47: ('i32.ne', None, 2, 1, 'sign-agnostic compare unequal'),
    0x48: ('i32.lt_s', None, 2, 1, 'signed less than'),
    0x49: ('i32.lt_u', None, 2, 1, 'unsigned less than'),
    0x4a: ('i32.gt_s', None, 2, 1, 'signed greater than'),
    0x4b: ('i32.gt_u', None, 2, 1, 'unsigned greater than'),
    0x4c: ('i32.le_s', None, 2, 1, 'signed less than or equal'),
    0x4d: ('i32.le_u', None, 2, 1, 'unsigned less than or equal'),
    0x4e: ('i32.ge_s', None, 2, 1, 'signed greater than or equal'),
    0x4f: ('i32.ge_u', None, 2, 1, 'unsigned greater than or equal'),

    0x50: ('i64.eqz', None, 1, 1, 'compare equal to zero (return 1 if operand is zero, 0 otherwise)'),
    0x51: ('i64.eq', None, 2, 1, 'sign-agnostic compare equal'),
    0x52: ('i64.ne', None, 2, 1, 'sign-agnostic compare unequal'),
    0x53: ('i64.lt_s', None, 2, 1, 'signed less than'),
    0x54: ('i64.lt_u', None, 2, 1, 'unsigned less than'),
    0x55: ('i64.gt_s', None, 2, 1, 'signed greater than'),
    0x56: ('i64.gt_u', None, 2, 1, 'unsigned greater than'),
    0x57: ('i64.le_s', None, 2, 1, 'signed less than or equal'),
    0x58: ('i64.le_u', None, 2, 1, 'unsigned less than or equal'),
    0x59: ('i64.ge_s', None, 2, 1, 'signed greater than or equal'),
    0x5a: ('i64.ge_u', None, 2, 1, 'unsigned greater than or equal'),

    0x5b: ('f32.eq', None, 2, 1, 'compare ordered and equal'),
    0x5c: ('f32.ne', None, 2, 1, 'compare unordered or unequal'),
    0x5d: ('f32.lt', None, 2, 1, 'compare ordered and less than'),
    0x5e: ('f32.gt', None, 2, 1, 'compare ordered and less than or equal'),
    0x5f: ('f32.le', None, 2, 1, 'compare ordered and greater than'),
    0x60: ('f32.ge', None, 2, 1, 'compare ordered and greater than or equal'),

    0x61: ('f64.eq', None, 2, 1, 'compare ordered and equal'),
    0x62: ('f64.ne', None, 2, 1, 'compare unordered or unequal'),
    0x63: ('f64.lt', None, 2, 1, 'compare ordered and less than'),
    0x64: ('f64.gt', None, 2, 1, 'compare ordered and less than or equal'),
    0x65: ('f64.le', None, 2, 1, 'compare ordered and greater than'),
    0x66: ('f64.ge', None, 2, 1, 'compare ordered and greater than or equal'),

    0x67: ('i32.clz', None, 1, 1, 'sign-agnostic count leading zero bits (All zero bits are considered leading if the value is zero)'),
    0x68: ('i32.ctz', None, 1, 1, 'sign-agnostic count trailing zero bits (All zero bits are considered trailing if the value is zero)'),
    0x69: ('i32.popcnt', None, 1, 1, 'sign-agnostic count number of one bits'),
    0x6a: ('i32.add', None, 2, 1, 'sign-agnostic addition'),
    0x6b: ('i32.sub', None, 2, 1, 'sign-agnostic subtraction'),
    0x6c: ('i32.mul', None, 2, 1, 'sign-agnostic multiplication (lower 32-bits)'),
    0x6d: ('i32.div_s', None, 2, 1, 'signed division (result is truncated toward zero)'),
    0x6e: ('i32.div_u', None, 2, 1, 'unsigned division (result is floored)'),
    0x6f: ('i32.rem_s', None, 2, 1, 'signed remainder (result has the sign of the dividend)'),
    0x70: ('i32.rem_u', None, 2, 1, 'unsigned remainder'),

    0x71: ('i32.and', None, 2, 1, 'sign-agnostic bitwise and'),
    0x72: ('i32.or', None, 2, 1, 'sign-agnostic bitwise inclusive or'),
    0x73: ('i32.xor', None, 2, 1, 'sign-agnostic bitwise exclusive or'),
    0x74: ('i32.shl', None, 2, 1, 'sign-agnostic shift left'),
    0x75: ('i32.shr_s', None, 2, 1, 'sign-replicating (arithmetic) shift right'),
    0x76: ('i32.shr_u', None, 2, 1, 'zero-replicating (logical) shift right'),
    0x77: ('i32.rotl', None, 2, 1, 'sign-agnostic rotate left'),
    0x78: ('i32.rotr', None, 2, 1, 'sign-agnostic rotate right'),

    0x79: ('i64.clz', None, 1, 1, 'sign-agnostic count leading zero bits (All zero bits are considered leading if the value is zero)'),
    0x7a: ('i64.ctz', None, 1, 1, 'sign-agnostic count trailing zero bits (All zero bits are considered trailing if the value is zero)'),
    0x7b: ('i64.popcnt', None, 1, 1, 'sign-agnostic count number of one bits'),
    0x7c: ('i64.add', None, 2, 1, 'sign-agnostic addition'),
    0x7d: ('i64.sub', None, 2, 1, 'sign-agnostic subtraction'),
    0x7e: ('i64.mul', None, 2, 1, 'sign-agnostic multiplication (lower 32-bits)'),
    0x7f: ('i64.div_s', None, 2, 1, 'signed division (result is truncated toward zero)'),
    0x80: ('i64.div_u', None, 2, 1, 'unsigned division (result is floored)'),
    0x81: ('i64.rem_s', None, 2, 1, 'signed remainder (result has the sign of the dividend)'),
    0x82: ('i64.rem_u', None, 2, 1, 'unsigned remainder'),

    0x83: ('i64.and', None, 2, 1, 'sign-agnostic bitwise and'),
    0x84: ('i64.or', None, 2, 1, 'sign-agnostic bitwise inclusive or'),
    0x85: ('i64.xor', None, 2, 1, 'sign-agnostic bitwise exclusive or'),
    0x86: ('i64.shl', None, 2, 1, 'sign-agnostic shift left'),
    0x87: ('i64.shr_s', None, 2, 1, 'sign-replicating (arithmetic) shift right'),
    0x88: ('i64.shr_u', None, 2, 1, 'zero-replicating (logical) shift right'),
    0x89: ('i64.rotl', None, 2, 1, 'sign-agnostic rotate left'),
    0x8a: ('i64.rotr', None, 2, 1, 'sign-agnostic rotate right'),

    0x8b: ('f32.abs', None, 1, 1, 'absolute value'),
    0x8c: ('f32.neg', None, 1, 1, 'negation'),
    0x8d: ('f32.ceil', None, 2, 1, 'ceiling operator'),
    0x8e: ('f32.floor', None, 2, 1, 'floor operator'),
    0x8f: ('f32.trunc', None, 2, 1, 'round to nearest integer towards zero'),
    0x90: ('f32.nearest', None, 2, 1, 'round to nearest integer, ties to even'),
    0x91: ('f32.sqrt', None, 2, 1, 'square root'),
    0x92: ('f32.add', None, 2, 1, 'addition'),
    0x93: ('f32.sub', None, 2, 1, 'subtraction'),
    0x94: ('f32.mul', None, 2, 1, 'multiplication'),
    0x95: ('f32.div', None, 2, 1, 'division'),
    0x96: ('f32.min', None, 2, 1, 'minimum (binary operator); if either operand is NaN, returns NaN'),
    0x97: ('f32.max', None, 2, 1, 'maximum (binary operator); if either operand is NaN, returns NaN'),
    0x98: ('f32.copysign', None, 2, 1, 'copysign'),

    0x99: ('f64.abs', None, 1, 1, 'absolute value'),
    0x9a: ('f64.neg', None, 1, 1, 'negation'),
    0x9b: ('f64.ceil', None, 2, 1, 'ceiling operator'),
    0x9c: ('f64.floor', None, 2, 1, 'floor operator'),
    0x9d: ('f64.trunc', None, 2, 1, 'round to nearest integer towards zero'),
    0x9e: ('f64.nearest', None, 2, 1, 'round to nearest integer, ties to even'),
    0x9f: ('f64.sqrt', None, 2, 1, 'square root'),
    0xa0: ('f64.add', None, 2, 1, 'addition'),
    0xa1: ('f64.sub', None, 2, 1, 'subtraction'),
    0xa2: ('f64.mul', None, 2, 1, 'multiplication'),
    0xa3: ('f64.div', None, 2, 1, 'division'),
    0xa4: ('f64.min', None, 2, 1, 'minimum (binary operator); if either operand is NaN, returns NaN'),
    0xa5: ('f64.max', None, 2, 1, 'maximum (binary operator); if either operand is NaN, returns NaN'),
    0xa6: ('f64.copysign', None, 2, 1, 'copysign'),

    0xa7: ('i32.wrap/i64', None, 1, 1, 'wrap a 64-bit integer to a 32-bit integer'),
    0xa8: ('i32.trunc_s/f32', None, 1, 1, 'truncate a 32-bit float to a signed 32-bit integer'),
    0xa9: ('i32.trunc_u/f32', None, 1, 1, 'truncate a 32-bit float to an unsigned 32-bit integer'),
    0xaa: ('i32.trunc_s/f64', None, 1, 1, 'truncate a 64-bit float to a signed 32-bit integer'),
    0xab: ('i32.trunc_u/f64', None, 1, 1, 'truncate a 64-bit float to an unsigned 32-bit integer'),
    0xac: ('i64.extend_s/i32', None, 1, 1, 'extend a signed 32-bit integer to a 64-bit integer'),
    0xad: ('i64.extend_u/i32', None, 1, 1, 'extend an unsigned 32-bit integer to a 64-bit integer'),
    0xae: ('i64.trunc_s/f32', None, 1, 1, 'truncate a 32-bit float to a signed 64-bit integer'),
    0xaf: ('i64.trunc_u/f32', None, 1, 1, 'truncate a 32-bit float to an unsigned 64-bit integer'),
    0xb0: ('i64.trunc_s/f64', None, 1, 1, 'truncate a 64-bit float to a signed 64-bit integer'),
    0xb1: ('i64.trunc_u/f64', None, 1, 1, 'truncate a 64-bit float to an unsigned 64-bit integer'),
    0xb2: ('f32.convert_s/i32', None, 1, 1, 'convert a signed 32-bit integer to a 32-bit float'),
    0xb3: ('f32.convert_u/i32', None, 1, 1, 'convert an unsigned 32-bit integer to a 32-bit float'),
    0xb4: ('f32.convert_s/i64', None, 1, 1, 'convert a signed 64-bit integer to a 32-bit float'),
    0xb5: ('f32.convert_u/i64', None, 1, 1, 'convert an unsigned 64-bit integer to a 32-bit float'),
    0xb6: ('f32.demote/f64', None, 1, 1, 'demote a 64-bit float to a 32-bit float'),
    0xb7: ('f64.convert_s/i32', None, 1, 1, 'convert a signed 32-bit integer to a 64-bit float'),
    0xb8: ('f64.convert_u/i32', None, 1, 1, 'convert an unsigned 32-bit integer to a 64-bit float'),
    0xb9: ('f64.convert_s/i64', None, 1, 1, 'convert a signed 64-bit integer to a 64-bit float'),
    0xba: ('f64.convert_u/i64', None, 1, 1, 'convert an unsigned 64-bit integer to a 64-bit float'),
    0xbb: ('f64.promote/f32', None, 1, 1, 'promote a 32-bit float to a 64-bit float'),
    0xbc: ('i32.reinterpret/f32', None, 1, 1, 'reinterpret the bits of a 32-bit float as a 32-bit integer'),
    0xbd: ('i64.reinterpret/f64', None, 1, 1, 'reinterpret the bits of a 64-bit float as a 64-bit integer'),
    0xbe: ('f32.reinterpret/i32', None, 1, 1, 'reinterpret the bits of a 32-bit integer as a 32-bit float'),
    0xbf: ('f64.reinterpret/i64', None, 1, 1, 'reinterpret the bits of a 64-bit integer as a 64-bit float'),
}


class Wasm(object):
    """Wasm bytecode."""

    def __init__(self):
        self.table = _table
        self.reverse_table = self._get_reverse_table()

    def _get_reverse_table(self):
        """Build an internal table used in the assembler."""
        # opcode:(mnemonic/name, imm_struct, pops, pushes, description)
        reverse_table = {}
        for (opcode, (mnemonic, imm_struct,
                      pops, pushes, description)) in self.table.items():
            reverse_table[mnemonic] = opcode, mnemonic, imm_struct, pops, pushes, description
        return reverse_table



class CFG(object):

    def __init__(self):
        """ TODO """
        raise NotImplementedError

    def visualize(self):
        """ TODO """
        raise NotImplementedError

    def visualize_call_flow(self):
        """ TODO """
        raise NotImplementedError

    def visualize_instrs_per_funcs(self):
        """ TODO """
        raise NotImplementedError



from octopus.analysis.cfg import CFG
from octopus.analysis.graph import CFGGraph
from octopus.core.function import Function
from octopus.core.basicblock import BasicBlock

from octopus.arch.evm.disassembler import EvmDisassembler

import json
import os

from logging import getLogger
logging = getLogger(__name__)


SIGNATURE_FILE_PATH = '/signatures.json'


def enum_func_static(instructions):

    functions = list()

    # first function is *usually* the function dispatcher
    function = Function(start_offset=0,
                        start_instr=instructions[0],
                        name='Dispatcher',
                        prefered_name='Dispatcher')
    functions.append(function)

    # parse the instructions and create Function object
    for inst in instructions:
        try:
            # PUSH4 are used to push the function signature on the stack
            if inst.name == 'PUSH4':
                index = instructions.index(inst)
                list_inst = instructions[index:index + 4]
                push4, eq, push, jumpi = list_inst[0], list_inst[1], list_inst[2], list_inst[3]

                # check if this basicblock test function signature
                if eq.name == 'EQ' and push.name in ['PUSH1', 'PUSH2'] and jumpi.name == 'JUMPI':
                    xref = int.from_bytes(push.operand, byteorder='big')
                    sign = int.from_bytes(push4.operand, byteorder='big')
                    name = 'func_%x' % sign
                    prefered_name = find_signature(sign)

                    # find instr with offset == xref
                    begin_function = next(filter(lambda i: i.offset == xref, instructions))
                    # create new function
                    function = Function(xref,
                                        start_instr=begin_function,
                                        name=name,
                                        prefered_name=prefered_name)

                    functions.append(function)
        except:
            logging.warning('enum_func_static Exception')
            pass
    return functions


def find_signature(sign):
    path = os.path.dirname(os.path.realpath(__file__)) + SIGNATURE_FILE_PATH
    with open(path) as data_file:
        data = json.load(data_file)

    list_name = [name for name, hexa in data.items() if hexa == str(hex(sign))]

    if len(list_name) > 1:
        logging.warning('function signatures collision: %s', list_name)
        return '_or_'.join(list_name)
    elif list_name:
        return list_name[0]
    else:
        return None


def enum_blocks_static(instructions):

    """
    Return a list of basicblock after
    statically parsing given instructions
    """

    basicblocks = list()
    index = 0

    # create the first block
    new_block = False
    end_block = False
    block = BasicBlock(instructions[0].offset,
                       instructions[0],
                       name='block_%x' % instructions[0].offset)

    for inst in instructions:
        if new_block:
            block = BasicBlock(inst.offset,
                               inst,
                               name='block_%x' % inst.offset)
            new_block = False

        # add current instruction to the basicblock
        block.instructions.append(inst)

        # absolute JUMP
        if inst.is_branch_unconditional:
            new_block = True

        # conditionnal JUMPI
        elif inst.is_branch_conditional:
            new_block = True

        # Halt instruction : RETURN, STOP, ...
        elif inst.is_halt:  # and inst != instructions[-1]:
            new_block = True

        # just falls to the next instruction
        elif inst != instructions[-1] and \
                instructions[index + 1].name == 'JUMPDEST':
            new_block = True

        # last instruction of the entire bytecode
        elif inst == instructions[-1]:
            end_block = True

        if new_block or end_block:
            block.end_offset = inst.offset_end
            block.end_instr = inst
            basicblocks.append(block)
            new_block = True
            end_block = False

        index += 1

    return basicblocks

'''
def enum_calls(instructions):
    for inst in instructions:
        # inst is a call instruction and ssa repr is not null
        if inst.is_call and inst.ssa:
            to_addr = inst.ssa.args[1]
            print('%s to %s' % (inst.name, str(to_addr)))
'''


class EvmCFG(CFG):

    def __init__(self, bytecode=None, analysis='dynamic'):
        """ TODO """

        self.bytecode = bytecode
        self.disasm = EvmDisassembler(self.bytecode)
        self.instructions = self.disasm.disassemble()
        self.analysis = analysis

        self.basicblocks = list()
        self.functions = list()
        self.edges = list()

        if self.analysis == 'dynamic':
            self.run_dynamic_analysis()
        elif self.analysis == 'static':
            self.run_static_analysis()

    def run_static_analysis(self):
        self.functions = enum_func_static(self.instructions)
        self.basicblocks = enum_blocks_static(self.instructions)

    def run_dynamic_analysis(self):

        from octopus.platforms.ETH.emulator import EthereumSSAEngine

        emul = EthereumSSAEngine(self.bytecode)
        emul.emulate()
        self.functions = emul.functions
        self.basicblocks = emul.basicblocks
        self.edges = emul.edges

    def visualize(self, simplify=False, ssa=False):
        """Visualize the cfg
        used CFGGraph
        equivalent to:
            graph = CFGGraph(cfg)
            graph.view_functions()
        """
        graph = CFGGraph(self)
        graph.view(simplify=simplify, ssa=ssa)

    def __str__(self):
        line = ("length functions = %d\n" % len(self.functions))
        line += ("length basicblocks = %d\n" % len(self.basicblocks))
        line += ("length instructions = %d\n" % len(self.instructions))
        line += ("length edges = %d\n" % len(self.edges))
        return line



from octopus.engine.emulator import EmulatorEngine
from octopus.arch.wasm.cfg import WasmCFG
from octopus.arch.wasm.vmstate import WasmVMstate
from octopus.core.ssa import SSA, SSA_TYPE_FUNCTION, SSA_TYPE_CONSTANT

from octopus.arch.wasm.format import (format_func_name,
                                      format_bb_name)

import copy

from logging import getLogger
logging = getLogger(__name__)

# =======================================
# #         WASM Emulator               #
# =======================================


class WasmSSAEmulatorEngine(EmulatorEngine):

    def __init__(self, bytecode, function_name, ssa=True):

        self.ssa = ssa

        # retrive instructions, basicblocks & functions statically
        self.cfg = WasmCFG(bytecode)
        self.ana = self.cfg.analyzer

        self.functions = self.cfg.functions
        if function_name not in [x.name for x in self.functions]:
            raise Exception('function_name not in this module - available: %s', self.ana.func_prototypes)

        self.current_function = None

        self.basicblock_per_instr = dict()
        self.current_basicblock = None

        self.current_function = self.cfg.get_function(function_name)
        self.current_f_instructions = self.current_function.instructions
        self.reverse_instructions = {k: v for k, v in enumerate(self.current_f_instructions)}
        self.current_f_basicblocks = self.current_function.basicblocks

        # create dict with:
        # * key = instruction offset
        # * value = basicblock reference
        # easy to get the corresponding basicblock per instr now
        for bb in self.current_f_basicblocks:
            for intr in bb.instructions:
                self.basicblock_per_instr[intr.offset] = bb

        # connection between basicblocks
        # will be generate dynamically by the Emulator

        self.states = dict()
        self.states_total = 0
        self.ssa_counter = 0

        logging.warning("=======================================")
        logging.warning("#      Wasm Emulator Engine")
        logging.warning("#      class: %s", self.__class__.__name__)
        logging.warning("=======================================")
        #for f in self.functions:
        logging.warning("[+] current_function detected - %x: %s/%s",
                        self.current_function.start_offset,
                        self.current_function.name,
                        self.current_function.prefered_name)

    def emulate(self, state=WasmVMstate(), depth=0):

        #  create fake stack for tests
        state.symbolic_stack = list(range(1000))

        # get current instruction
        instr = self.reverse_instructions[state.pc]

        # create the first basicblock of this branch
        # print('%d : %s' % (instr.offset, instr.name))
        self.current_basicblock = self.basicblock_per_instr[instr.offset]

        # beginning of a function
        #if instr in self.functions_start_instr:
            # retrive matching function
        #    self.current_function = next(filter(lambda f: f.start_instr == instr, self.functions))
            # self.ssa_counter = 0
        logging.warning("[+] Entering function - %x: %s",
                        self.current_function.start_offset,
                        self.current_function.prefered_name)

        # associate function to basicblock
        # TODO: create list of function_name
        self.current_basicblock.function_name = self.current_function.prefered_name
        # associate basicblock to function
        self.current_function.basicblocks.append(self.current_basicblock)

        # halt variable use to catch ending branch
        halt = False
        while not halt:

            # get current instruction
            instr = self.reverse_instructions[state.pc]

            # get current basicblock
            self.current_basicblock = self.basicblock_per_instr[instr.offset]

            # add this instruction to his functions
            # TODO: verify if it's not useless for ethereum
            #self.current_function.instructions.append(instr)

            # Save instruction and state
            state.instr = instr
            self.states[self.states_total] = state
            state = copy.deepcopy(state)
            self.states_total += 1
            state.pc += 1

            # execute single instruction
            halt = self.emulate_one_instruction(instr, state, depth)
            state.instructions_visited.append(instr.offset)
            #state.instructions_visited[instr.offset] = instr.offset

        logging.warning("[X] Returning from basicblock %s", self.current_basicblock.name)

        # automatic remove duplicated edges
        #self.edges = list(set(self.edges))
    '''
    is_control
    is_parametric
    is_variable
    is_memory
    is_constant
    is_logical_i32
    is_logical_i64
    is_logical_f32
    is_logical_f64
    is_arithmetic_i32
    is_bitwise_i32
    is_arithmetic_i64
    is_bitwise_i64
    is_arithmetic_f32
    is_arithmetic_f64
    is_conversion
    '''

    def emulate_one_instruction(self, instr, state, depth):

        halt = False

        logging.warning('--')
        logging.warning('stack %s' % state.ssa_stack)
        logging.warning('instr %s' % instr.name)
        logging.warning('operand %s' % instr.operand)
        logging.warning('xref %s' % instr.xref)

        if instr.is_control:
            halt = self.emul_control_instr(instr, state, depth)

        elif instr.is_parametric:
            halt = self.emul_parametric_instr(instr, state)

        elif instr.is_variable:
            halt = self.emul_variable_instr(instr, state)

        elif instr.is_memory:
            halt = self.emul_memory_instr(instr, state)

        elif instr.is_constant:
            halt = self.emul_constant_instr(instr, state)

        elif instr.is_logical_i32:
            halt = self.emul_logical_i32_instr(instr, state)

        elif instr.is_logical_i64:
            halt = self.emul_logical_i64_instr(instr, state)

        elif instr.is_logical_f32:
            halt = self.emul_logical_f32_instr(instr, state)

        elif instr.is_logical_f64:
            halt = self.emul_logical_f64_instr(instr, state)

        elif instr.is_arithmetic_i32:
            halt = self.emul_arithmetic_i32_instr(instr, state)

        elif instr.is_bitwise_i32:
            halt = self.emul_bitwise_i32_instr(instr, state)

        elif instr.is_arithmetic_i64:
            halt = self.emul_arithmetic_i64_instr(instr, state)

        elif instr.is_bitwise_i64:
            halt = self.emul_bitwise_i64_instr(instr, state)

        elif instr.is_arithmetic_f32:
            halt = self.emul_arithmetic_f32_instr(instr, state)

        elif instr.is_arithmetic_f64:
            halt = self.emul_arithmetic_f64_instr(instr, state)

        elif instr.is_conversion:
            halt = self.emul_conversion_instr(instr, state)

        # UNKNOWN INSTRUCTION
        else:
            logging.warning('UNKNOWN = ' + instr.name)

        return halt

    def emul_control_instr(self, instr, state, depth):
        halt = False
        if instr.name == 'unreachable':
            instr.ssa = SSA(method_name=instr.name)
            halt = True
        elif instr.name in ['nop', 'block', 'loop', 'else']:
            instr.ssa = SSA(method_name=instr.name)
        elif instr.name == 'if':
            arg = [state.ssa_stack.pop()]
            instr.ssa = SSA(method_name=instr.name, args=arg)
            # TODO branch if
            # inst + 1 == true block
            # need to find offset false block using edges or basicblocks list
            logging.warning('SSA: branch if not yet supported')
        elif instr.name == 'end':
            instr.ssa = SSA(method_name=instr.name)
            # check if it's the last instructions of the function
            if instr.offset == self.current_f_instructions[-1].offset:
                logging.warning("[X] break %s" % instr.name)
                halt = True
        elif instr.name == 'br':
            instr.ssa = SSA(method_name=instr.name)
            jump_addr = instr.xref

            # get instruction with this value as offset
            target = next(filter(lambda element: element.offset == jump_addr[0], self.current_f_instructions))

            if target.offset not in state.instructions_visited:
                # condition are True
                logging.warning('[X] follow br branch offset 0x%x' % (target.offset))
                new_state = copy.deepcopy(state)
                new_state.pc = self.current_f_instructions.index(target)
                # follow the br
                self.emulate(new_state, depth=depth + 1)
            else:
                logging.warning('[X] Loop detected, skipping br 0x%x' % jump_addr[0])
                halt = True
            halt = True

        elif instr.name == 'br_if':
            arg = [state.ssa_stack.pop()]
            instr.ssa = SSA(method_name=instr.name, args=arg)

            logging.warning('[X] follow br_if default branch offset 0x%x' % (instr.offset_end + 1))
            new_state = copy.deepcopy(state)

            self.emulate(new_state, depth=depth + 1)
            # after we return from emul - restore current_basicblock
            self.current_basicblock = self.basicblock_per_instr[instr.offset]

            jump_addr = instr.xref
            # get instruction with this value as offset
            target = next(filter(lambda element: element.offset == jump_addr[0], self.current_f_instructions))

            if target.offset not in state.instructions_visited:
                # condition are True
                logging.warning('[X] follow br_if branch offset 0x%x' % (target.offset))
                new_state = copy.deepcopy(state)
                new_state.pc = self.current_f_instructions.index(target)

                # follow the br_if
                self.emulate(new_state, depth=depth + 1)

            else:
                logging.warning('[X] Loop detected, skipping br_if 0x%x' % jump_addr[0])
                halt = True
            halt = True
        elif instr.name == 'br_table':
            arg = [state.ssa_stack.pop()]
            instr.ssa = SSA(method_name=instr.name, args=arg)
            # TODO branch br_table
            logging.warning('SSA: branch br_table not yet supported')
        elif instr.name == 'return':
            arg = [state.ssa_stack.pop()]
            instr.ssa = SSA(method_name=instr.name, args=arg)
            halt = True
        elif instr.name == 'call':
            f_offset = int.from_bytes(instr.operand, 'big')
            target_func = self.ana.func_prototypes[f_offset]
            name, param_str, return_str, f_type = target_func
            # format_func_name()
            instr.ssa = SSA(method_name=instr.name + '_to_' + name)
            if param_str:
                num_arg = len(param_str.split(' '))
                #print(hex(state.ssa_stack[-1].offset))
                arg = [state.ssa_stack.pop() for x in range(1, num_arg+1)]
                instr.ssa.args = arg
            if return_str:
                instr.ssa.new_assignement = self.ssa_counter
                state.ssa_stack.append(instr)
                self.ssa_counter += 1
        elif instr.name == 'call_indirect':
            arg = [state.ssa_stack.pop()]
            ''' # issue when table is imported
            # arg is constant
            if arg[0].ssa.instr_type == SSA_TYPE_CONSTANT:
                f_offset = int.from_bytes(instr.operand, 'big')
                target_func = self.ana.func_prototypes[f_offset]

                name, param_str, return_str, f_type = target_func
                # format_func_name()
                instr.ssa = SSA(method_name=instr.name + '_to_' + name)
                if param_str:
                    num_arg = len(param_str.split(' '))
                    print(hex(state.ssa_stack[-1].offset))
                    arg = [state.ssa_stack.pop() for x in range(1, num_arg+1)]
                    instr.ssa.args = arg
                if return_str:
                    instr.ssa.new_assignement = self.ssa_counter
                    state.ssa_stack.append(instr)
                    self.ssa_counter += 1
            else:
            '''
            instr.ssa = SSA(method_name=instr.name, args=arg)
            # test if arg is constant --> do like call
            # else - stay like that
        return halt

    def emul_parametric_instr(self, instr, state):
        if instr.name == 'drop':
            state.ssa_stack.pop()
            instr.ssa = SSA(method_name=instr.name)
        else:  # select
            arg = [state.ssa_stack.pop(), state.ssa_stack.pop(), state.ssa_stack.pop()]
            instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
            state.ssa_stack.append(instr)
            self.ssa_counter += 1
        return False

    def emul_variable_instr(self, instr, state):
        if instr.name in ['get_local', 'get_global']:
            instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name)
            state.ssa_stack.append(instr)
            self.ssa_counter += 1
        elif instr.name in ['set_local', 'set_global']:
            state.ssa_stack.pop()
            instr.ssa = SSA(method_name=instr.name)
        elif instr.name == 'tee_local':
            state.ssa_stack.append(state.ssa_stack[-1])
            self.ssa_counter += 1
        return False

    def emul_memory_instr(self, instr, state):
        # load
        if 'load' in instr.name:
            arg = [state.ssa_stack.pop()]
            instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
            state.ssa_stack.append(instr)
            self.ssa_counter += 1

        elif 'store' in instr.name:
            arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
            instr.ssa = SSA(method_name=instr.name, args=arg)

        elif instr.name == 'current_memory':
            instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name)
            state.ssa_stack.append(instr)
            self.ssa_counter += 1
        else:
            instr.ssa = SSA(method_name=instr.name)
        return False

    def emul_constant_instr(self, instr, state):
        op = int.from_bytes(instr.operand, byteorder='big')
        instr.ssa = SSA(self.ssa_counter, instr.name,
                        op,
                        instr_type=SSA_TYPE_CONSTANT)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1
        return False

    def emul_logical_i32_instr(self, instr, state):
        if instr.name == 'i32.eqz':
            arg = [state.ssa_stack.pop()]
        else:
            arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_logical_i64_instr(self, instr, state):
        if instr.name == 'i64.eqz':
            arg = [state.ssa_stack.pop()]
        else:
            arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_logical_f32_instr(self, instr, state):
        arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_logical_f64_instr(self, instr, state):
        arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_arithmetic_i32_instr(self, instr, state):
        if instr.name in ['i32.clz', 'i32.ctz', 'i32.popcnt']:
            arg = [state.ssa_stack.pop()]
        else:
            arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_bitwise_i32_instr(self, instr, state):
        arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_arithmetic_i64_instr(self, instr, state):
        if instr.name in ['i64.clz', 'i64.ctz', 'i64.popcnt']:
            arg = [state.ssa_stack.pop()]
        else:
            arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_bitwise_i64_instr(self, instr, state):
        arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_arithmetic_f32_instr(self, instr, state):
        if instr.name in ['f32.abs', 'f32.neg']:
            arg = [state.ssa_stack.pop()]
        else:
            arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_arithmetic_f64_instr(self, instr, state):
        if instr.name in ['f64.abs', 'f64.neg']:
            arg = [state.ssa_stack.pop()]
        else:
            arg = [state.ssa_stack.pop(), state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False

    def emul_conversion_instr(self, instr, state):
        arg = [state.ssa_stack.pop()]
        instr.ssa = SSA(new_assignement=self.ssa_counter, method_name=instr.name, args=arg)
        state.ssa_stack.append(instr)
        self.ssa_counter += 1

        return False



# Global Ethereum-specific constant

DEFAULT_GAS_PER_TX = 90000
DEFAULT_GAS_PRICE = 50 * 10**9  # 50 gwei

BLOCK_TAG_EARLIEST = 'earliest'
BLOCK_TAG_LATEST = 'latest'
BLOCK_TAG_PENDING = 'pending'
BLOCK_TAGS = (
    BLOCK_TAG_EARLIEST,
    BLOCK_TAG_LATEST,
    BLOCK_TAG_PENDING,
)


TT256 = 2 ** 256
TT256M1 = 2 ** 256 - 1
TT255 = 2 ** 255
